unit WndWrapper;

{$mode ObjFPC}{$H+}

interface

uses
  xlib,
  x,
  xutil,
  ctypes,
  objects,
  // Diese beiden Units müssen am Ende stehen sonst geht die TStringList nicht.
  Classes, SysUtils;

// Diese Struktur und die nachfolgenden Konstanten sind der CustomDrawn-Bibliothek
// entnommen genauer gesagt aus:
// lazarus/lcl/interfaces/customdrawn/customdrawnwsforms_x11.inc
// Sie werden dafür benötigt um die Fensterdekoration zu entfernen.
type
  PMotifWmHints = ^TMotifWmHints;
  TMotifWmHints = packed record
    Flags, Functions, Decorations: LongWord;
    InputMode: LongInt;
    Status: LongWord;
  end;

const
  MWM_HINTS_FUNCTIONS = 1;
  MWM_HINTS_DECORATIONS = 2;
  FuncAll = 1;
  FuncResize = 2;
  FuncMove = 4;
  FuncMinimize = 8;
  FuncMaximize = 16;
  FuncClose = 32;
  DecorAll = 1;
  DecorBorder = 2;
  DecorResizeH = 4;
  DecorTitle = 8;
  DecorMenu = 16;
  DecorMinimize = 32;
  DecorMaximize = 64;

type
  // TWndWrapper ist der Versuch die von mir benötigten Funktionen in einer Klasse
  // zu kapseln.
  // Diese Klasse dient zu Testzwecken und als Basis eine vernünftige Struktur zu
  // entwickeln.
  TWndWrapper = class
  private
    // Diese Variablen werden für fast alle XLib-Funktionen benötigt
    fDisplay : PDisplay;
    fScreen : cint;
    fRootWnd : TWindow;
    // Das Window der Hauptanwendung
    fMainWnd : TWindow;
    // Liste der Fenster in Form von Name=Window (Cardinal)
    fWList : TStringList;
    procedure ConnectServer;
    procedure GetDefaultScreen;
    procedure GetRootWindow;
    // Funktionen von theo übernommen
    function GetWindowName(Window: TWindow): string;
    procedure EnumerateTopLevelWindows;
  public
    constructor Create(aOwnerName: string);
    destructor Destroy; override;
    function SelectWindow(APartName: string): TWindow;
    procedure MoveWndTo(aWnd: TWindow; aX, aY: Integer);
    procedure ResizeWindow(aWnd: TWindow; Width, Height:integer);
    procedure MakeMainParent(aWnd: TWindow; aX, aY: Integer);
    procedure HideDecorations(aWnd: TWindow);
    procedure BringToFront(aWnd: TWindow);
    function GetAbsCoord(aWnd: TWindow) : TPoint;
    function GetAbsMainCoord(aPoint: TPoint): TPoint;
    procedure Minimize(aWnd: TWindow);
    procedure Normal(aWnd: TWindow);
    property WindowList : TStringList read fWList;
  end;

// Übernommen von theo
  TWindowArray = array[0..MaxListSize] of TWindow;
  PWindowArray = ^TWindowArray;

threadvar
  GotBadwindow: boolean;

threadvar
  OldHandler: TXErrorHandler;

function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: cardinal;
  Offset: cardinal = 0): boolean; overload;
function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: string;
  Offset: cardinal = 0): boolean; overload;
function GetState(Dpy: PDisplay; Win: TWindow): cardinal;

implementation

uses
  // Für ShowMessage zum debuggen
  Dialogs
  ;

// Gibt eine Cardinal-Eigenschaft in Value zurück (Code von theo)
function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: cardinal;
  Offset: cardinal): boolean;
var
  ActualTypeReturn: TAtom;
  ActualFormatReturn: longint;
  NItemsReturn, BytesAfterReturn: cardinal;
  Ptr: pbyte;
begin
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 1, longbool(0),
    AnyPropertyType, @ActualTypeReturn, @ActualFormatReturn, @NItemsReturn,
    @BytesAfterReturn, @Ptr) = Success;
  if Result then
    try
    {$message warn: ActualFormatReturn always 0 on 64bit}
      if (ActualTypeReturn = None) {or (ActualFormatReturn <> 32) } or
        not Assigned(Ptr) then
      begin
        Result := False;
      end;
      if Result then Value := PCardinal(Ptr)^;
    finally
      if Assigned(Ptr) then XFree(Ptr);
    end;
end;

// Gibt eine String-Eigenschaft in Value zurück (Code von theo)
function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: string;
  Offset: cardinal = 0): boolean;
var
  ActualTypeReturn: TAtom;
  ActualFormatReturn: longint;
  Length, NItemsReturn, BytesAfterReturn: cardinal;
  Ptr: pbyte;
begin
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 0, longbool(0),
    None, @ActualTypeReturn, @ActualFormatReturn, @NItemsReturn,
    @BytesAfterReturn, @Ptr) = Success;
  if Result and Assigned(Ptr) then
    XFree(Ptr);
  if (ActualTypeReturn = None) or (ActualFormatReturn <> 8) then Result := False;
  if Result then
    try
      Length := (BytesAfterReturn + 3) div 4;
      Result := XGetWindowProperty(Dpy, Win, Prop, Offset, Length,
        longbool(0), ActualTypeReturn, @ActualTypeReturn, @ActualFormatReturn,
        @NItemsReturn, @BytesAfterReturn, @Ptr) = Success;
      if (ActualTypeReturn = None) or (ActualFormatReturn <> 8) or not Assigned(Ptr) then
        Result := False;
      if Result then
        SetString(Value, PChar(Ptr), NItemsReturn);
    finally
      if Assigned(Ptr) then
        XFree(Ptr);
    end;
end;

// Gibt den Status zurück (Code von theo)
function GetState(Dpy: PDisplay; Win: TWindow): cardinal;
var
  A: TAtom;
begin
  A := XInternAtom(Dpy, 'WM_STATE', longbool(1));
  if A = None then
    Result := 0
  else if not GetProp(Dpy, Win, A, Result) then
    Result := 0;
end;

// Behandelt den Badwindow-Fehler (Code von theo)
function BadwindowHandler(para1: PDisplay; para2: PXErrorEvent): cint; cdecl;
begin
  if para2^.error_code <> BadWindow then
    Result := OldHandler(para1, para2)
  else
  begin
    GotBadwindow := True;
    Result := 0;
  end;
end;

// Konstruktor
// Ermittelt die allgemeinen Parameter die immer wieder benötigt werden wie
// fDisplay, fScreen, fRootWnd.
// Erzeugt die Fensterliste fWList und liest ihre Daten ein -> EnumerateTopLevelWindows
// fMainWnd wird aus der Fensterliste selektiert.
constructor TWndWrapper.Create(aOwnerName: string);
begin
  inherited Create;
  fWList := TStringList.Create;
  ConnectServer;
  if fDisplay <> nil then
  begin
    GetDefaultScreen;
    GetRootWindow;
    EnumerateTopLevelWindows;
    fMainWnd := SelectWindow(aOwnerName);
  end;
end;

// Gibt Speicher wieder frei.
destructor TWndWrapper.Destroy;
begin
  // Schliesst Verbindung zum Server
  if fDisplay <> nil then
    XCloseDisplay(fDisplay);
  fWList.Free;
  inherited Destroy;
end;

// Erstellt die Verbindung zum Server
procedure TWndWrapper.ConnectServer;
begin
  fDisplay := XOpenDisplay(nil);
  if (fDisplay = nil) then
    WriteLn('[ConnectServer] Kann das Diplay nicht oeffnen');
end;

// Sreen-Nr ermitteln
procedure TWndWrapper.GetDefaultScreen;
begin
  fScreen := DefaultScreen(fDisplay);
end;

// RootWindow ermitteln
procedure TWndWrapper.GetRootWindow;
begin
  fRootWnd := RootWindow(fDisplay, fScreen);
end;

// Ermittelt den Fenstertitel des angegebenen Window (Code von theo)
function TWndWrapper.GetWindowName(Window: TWindow): string;
var
  Stat: XLib.TStatus;
  text_prop: TXTextProperty;
  List: PPChar;
  Num: integer;
begin
  Result := '';
  OldHandler := XSetErrorHandler(@BadwindowHandler);
  GotBadwindow := False;
  stat := XGetWMName(fDisplay, Window, @text_prop);
  XSync(fDisplay, False);
  if GotBadwindow then
  begin
    XSetErrorHandler(OldHandler);
    exit;
  end;
  if ((stat = 0) or (text_prop.Value = nil) or (text_prop.nitems = 0)) then
  begin
    XSetErrorHandler(OldHandler);
    exit;
  end;
  stat := XmbTextPropertyToTextList(fDisplay, @text_prop, @List, @Num);
  if ((stat < Success) or (num = 0) or (list = nil)) then
  begin
    XSetErrorHandler(OldHandler);
    exit;
  end
  else
  begin
    XFree(text_prop.Value);
    if Num > 0 then Result := List^;
    XFreeStringList(list);
  end;
  XSetErrorHandler(OldHandler);
end;

// Erzeugt die Fensterliste (leicht angepasster Code von theo)
procedure TWndWrapper.EnumerateTopLevelWindows;
   procedure QueryChildren(Win: TWindow);
  var
    Root, Parent: TWindow;
    Children: PWindowArray;
    I, Count: cardinal;
    s : string;
  begin
    if XQueryTree(fDisplay, Win, @Root, @Parent, PPWindow(@Children), @Count) = 0 then Exit;
    if Count = 0 then Exit;
    try
      for I := 0 to Count - 1 do
      begin
        GotBadwindow := False;
        QueryChildren(Children^[I]);
        XSync(fDisplay, False);
        if (not GotBadwindow) and (GetState(fDisplay, Children^[I]) <> 0) then
        begin
          // Das weicht von theo's code ab
          s := GetWindowName(Children^[I])+'='+IntToStr(Children^[I]);
          fWList.Add(s);
        end;
      end;
    finally
      XFree(Children);
    end;
  end;
 begin
  try
    fWList.Clear;
    OldHandler := XSetErrorHandler(@BadwindowHandler);
    QueryChildren(fRootWnd);
  finally
    XSetErrorHandler(OldHandler);
  end;
end;

// Gibt das erste Window zurück, das APartName als Teil des Namens beinhaltet.
// 0 Wenn nicht gefunden.
function TWndWrapper.SelectWindow(APartName: string): TWindow;
Var
  i : Integer;
begin
  result := 0;
  for i := 0 to fWList.Count - 1 do
  begin
    if Pos(APartName, fWList.Names[i]) <> 0 then
    begin
      result := StrToInt(fWList.ValueFromIndex[i]);
      break;
    end;
  end;
end;

// Verschiebt das Fenster zur angegebenen Position
procedure TWndWrapper.MoveWndTo(aWnd: TWindow; aX, aY: Integer);
begin
  if ((fDisplay <> nil) and (AWnd <> 0)) then
  begin
    XMoveWindow(fDisplay, aWnd, aX, aY);
    XFlush(fDisplay);
  end;
end;

// Das Funktioniert nicht wie erhofft.
// Das Fenster verschwindet komplett, ist aber noch im Taskmanager gelistet.
// Wenn das jemand hinkriegt, wäre ich an der Lösung interessiert.
procedure TWndWrapper.MakeMainParent(aWnd: TWindow; aX, aY: Integer);
begin
  if ((fDisplay <> nil) and (AWnd <> 0)) then
  begin
    XReparentWindow(fDisplay, aWnd, fMainWnd, 0, 0);
//    sleep(100);
    XSync(fDisplay, False);
    //XFlush(fDisplay);
  end;
end;

// Entfernen der Fensterdekoration
procedure TWndWrapper.HideDecorations(aWnd: TWindow);
var
  a : TAtom;
  MwmH : TMotifWmHints;
begin
  a := XInternAtom(fDisplay, '_MOTIF_WM_HINTS', 0);
  if (a <> None) and (AWnd <> 0) then
  begin
    MwmH.Flags := MWM_HINTS_DECORATIONS;
    MwmH.Decorations := 0;    // Fixe Größe und Position durch Maus nicht zu ändern
    //MwmH.Decorations := DecorBorder; // Größe änderbar per Maus (Position indirekt über Größe änderbar)
    //MwmH.Decorations := DecorBorder or DecorTitle; // Keine Änderung zum normalen Fenster

    XChangeProperty(fDisplay, aWnd,
      a, a,
      32, PropModeReplace, Pointer(@MwmH), 5);
    XFlush(fDisplay);
  end;
end;

// Fenster nach vorne bringen
procedure TWndWrapper.BringToFront(aWnd: TWindow);
begin
  if ((fDisplay <> nil) and (AWnd <> 0)) then
  begin
    XRaiseWindow(fDisplay, aWnd);
    XFlush(fDisplay);
  end;
end;

// Gibt die absoluten Koordinaten des Fensters auf dem Display an.
// Hinweis: Die Fensterdekoration wird dabei nicht berücksichtigt. D.h. Wenn
// man sich die Fensterdekoration (Rahmen, Titelleiste, etc.) wegdenkt, dann
// ist oben links die Koordinate, die man erhält.
function TWndWrapper.GetAbsCoord(aWnd: TWindow): TPoint;
Var
  w : TWindow;
begin
  result.Create(-1, -1);
  if ((fDisplay <> nil) and (AWnd <> 0)) then
  begin
    XTranslateCoordinates(fDisplay, aWnd, fRootWnd, 0, 0, @(result.X), @(result.y), @w);
  end;
end;

// Hilfsfunktion ähnlich GetAbsCoord
// Es wird die absolute Position bezüglich eines Punktes innerhalb des Haupt-
// fensters berechnet. Danach kann ein anderes Fenster positioniert werden.
function TWndWrapper.GetAbsMainCoord(aPoint: TPoint): TPoint;
Var
  w : TWindow;
begin
  result.Create(-1, -1);
  if (fDisplay <> nil) then
  begin
    XTranslateCoordinates(fDisplay, fMainWnd, fRootWnd, aPoint.X, aPoint.Y, @(result.X), @(result.y), @w);
  end;
end;

// Ändern der Größe eines Fensters.
procedure TWndWrapper.ResizeWindow(aWnd: TWindow; Width, Height:integer);
begin
    XResizeWindow(fDisplay, aWnd, Width, Height);
    XFlush(fdisplay);
end;

// Fenster in die Taskleiste minimieren
procedure TWndWrapper.Minimize(aWnd: TWindow);
begin
 XIconifyWindow(fDisplay, aWnd, fScreen);
 XFlush(fdisplay);
end;

// Fenster an seiner letzen Position wieder darstellen.
procedure TWndWrapper.Normal(aWnd: TWindow);
begin
 XMapWindow(fDisplay, aWnd);
 XFlush(fdisplay);
end;

end.

